@@ -11,7 +11,7 @@ module Bundler
 
     AUTO_INSTALL_CMDS = %w[show binstubs outdated exec open console licenses clean].freeze
     PARSEABLE_COMMANDS = %w[check config help exec platform show version].freeze
-    EXTENSIONS = ["c", "rust"].freeze
+    EXTENSIONS = ["c", "rust", "go"].freeze
 
     COMMAND_ALIASES = {
       "check" => "c",
@@ -24,7 +24,7 @@ module Bundler
     }.freeze
 
     def self.start(*)
-      check_deprecated_ext_option(ARGV) if ARGV.include?("--ext")
+      check_invalid_ext_option(ARGV) if ARGV.include?("--ext")
 
       super
     ensure
@@ -59,17 +59,29 @@ module Bundler
     def initialize(*args)
       super
 
+      current_cmd = args.last[:current_command].name
+
       custom_gemfile = options[:gemfile] || Bundler.settings[:gemfile]
       if custom_gemfile && !custom_gemfile.empty?
         Bundler::SharedHelpers.set_env "BUNDLE_GEMFILE", File.expand_path(custom_gemfile)
-        Bundler.reset_settings_and_root!
+        reset_settings = true
       end
 
+      # lock --lockfile works differently than install --lockfile
+      unless current_cmd == "lock"
+        custom_lockfile = options[:lockfile] || ENV["BUNDLE_LOCKFILE"] || Bundler.settings[:lockfile]
+        if custom_lockfile && !custom_lockfile.empty?
+          Bundler::SharedHelpers.set_env "BUNDLE_LOCKFILE", File.expand_path(custom_lockfile)
+          reset_settings = true
+        end
+      end
+
+      Bundler.reset_settings_and_root! if reset_settings
+
       Bundler.auto_switch
 
       Bundler.settings.set_command_option_if_given :retry, options[:retry]
 
-      current_cmd = args.last[:current_command].name
       Bundler.auto_install if AUTO_INSTALL_CMDS.include?(current_cmd)
     rescue UnknownArgumentError => e
       raise InvalidOption, e.message
@@ -107,7 +119,30 @@ module Bundler
       shell.say
       self.class.send(:class_options_help, shell)
     end
-    default_task(Bundler.feature_flag.default_cli_command)
+
+    desc "install_or_cli_help", "Tries to run bundle install but prints a summary of bundler commands if there is no Gemfile", hide: true
+    def install_or_cli_help
+      invoke_other_command("install")
+    rescue GemfileNotFound => error
+      Bundler.ui.error error.message, wrap: true
+      invoke_other_command("cli_help")
+    end
+
+    def self.default_command(meth = nil)
+      return super if meth
+
+      unless Bundler.settings[:default_cli_command]
+        Bundler.ui.info <<-MSG
+          In a future version of Bundler, running `bundle` without argument will no longer run `bundle install`.
+          Instead, the `cli_help` command will be displayed. Please use `bundle install` explicitly for scripts like CI/CD.
+          You can use the future behavior now with `bundle config set default_cli_command cli_help --global`,
+          or you can continue to use the current behavior with `bundle config set default_cli_command install_or_cli_help --global`.
+          This message will be removed after a default_cli_command value is set.
+        MSG
+      end
+
+      Bundler.settings[:default_cli_command] || "install_or_cli_help"
+    end
 
     class_option "no-color", type: :boolean, desc: "Disable colorization in output"
     class_option "retry", type: :numeric, aliases: "-r", banner: "NUM",
@@ -143,7 +178,7 @@ module Bundler
     end
 
     def self.handle_no_command_error(command, has_namespace = $thor_runner)
-      if Bundler.feature_flag.plugins? && Bundler::Plugin.command?(command)
+      if Bundler.settings[:plugins] && Bundler::Plugin.command?(command)
         return Bundler::Plugin.exec_command(command, ARGV[1..-1])
       end
 
@@ -173,7 +208,7 @@ module Bundler
     D
     method_option "dry-run", type: :boolean, default: false, banner: "Lock the Gemfile"
     method_option "gemfile", type: :string, banner: "Use the specified gemfile instead of Gemfile"
-    method_option "path", type: :string, banner: "Specify a different path than the system default ($BUNDLE_PATH or $GEM_HOME).#{" Bundler will remember this value for future installs on this machine" unless Bundler.feature_flag.forget_cli_options?}"
+    method_option "path", type: :string, banner: "Specify a different path than the system default, namely, $BUNDLE_PATH or $GEM_HOME (removed)"
     def check
       remembered_flag_deprecation("path")
 
@@ -187,12 +222,11 @@ module Bundler
     long_desc <<-D
       Removes the given gems from the Gemfile while ensuring that the resulting Gemfile is still valid. If the gem is not found, Bundler prints a error message and if gem could not be removed due to any reason Bundler will display a warning.
     D
-    method_option "install", type: :boolean, banner: "Runs 'bundle install' after removing the gems from the Gemfile"
+    method_option "install", type: :boolean, banner: "Runs 'bundle install' after removing the gems from the Gemfile (removed)"
     def remove(*gems)
       if ARGV.include?("--install")
-        message = "The `--install` flag has been deprecated. `bundle install` is triggered by default."
         removed_message = "The `--install` flag has been removed. `bundle install` is triggered by default."
-        SharedHelpers.major_deprecation(2, message, removed_message: removed_message)
+        raise InvalidOption, removed_message
       end
 
       require_relative "cli/remove"
@@ -210,27 +244,29 @@ module Bundler
 
       If the bundle has already been installed, bundler will tell you so and then exit.
     D
-    method_option "binstubs", type: :string, lazy_default: "bin", banner: "Generate bin stubs for bundled gems to ./bin"
-    method_option "clean", type: :boolean, banner: "Run bundle clean automatically after install"
-    method_option "deployment", type: :boolean, banner: "Install using defaults tuned for deployment environments"
-    method_option "frozen", type: :boolean, banner: "Do not allow the Gemfile.lock to be updated after this install"
+    method_option "binstubs", type: :string, lazy_default: "bin", banner: "Generate bin stubs for bundled gems to ./bin (removed)"
+    method_option "clean", type: :boolean, banner: "Run bundle clean automatically after install (removed)"
+    method_option "deployment", type: :boolean, banner: "Install using defaults tuned for deployment environments (removed)"
+    method_option "frozen", type: :boolean, banner: "Do not allow the Gemfile.lock to be updated after this install (removed)"
     method_option "full-index", type: :boolean, banner: "Fall back to using the single-file index of all gems"
     method_option "gemfile", type: :string, banner: "Use the specified gemfile instead of Gemfile"
     method_option "jobs", aliases: "-j", type: :numeric, banner: "Specify the number of jobs to run in parallel"
     method_option "local", type: :boolean, banner: "Do not attempt to fetch gems remotely and use the gem cache instead"
+    method_option "lockfile", type: :string, banner: "Use the specified lockfile instead of the default."
     method_option "prefer-local", type: :boolean, banner: "Only attempt to fetch gems remotely if not present locally, even if newer versions are available remotely"
     method_option "no-cache", type: :boolean, banner: "Don't update the existing gem cache."
+    method_option "no-lock", type: :boolean, banner: "Don't create a lockfile."
     method_option "force", type: :boolean, aliases: "--redownload", banner: "Force reinstalling every gem, even if already installed"
-    method_option "no-prune", type: :boolean, banner: "Don't remove stale gems from the cache."
-    method_option "path", type: :string, banner: "Specify a different path than the system default ($BUNDLE_PATH or $GEM_HOME).#{" Bundler will remember this value for future installs on this machine" unless Bundler.feature_flag.forget_cli_options?}"
+    method_option "no-prune", type: :boolean, banner: "Don't remove stale gems from the cache (removed)."
+    method_option "path", type: :string, banner: "Specify a different path than the system default, namely, $BUNDLE_PATH or $GEM_HOME (removed)."
     method_option "quiet", type: :boolean, banner: "Only output warnings and errors."
-    method_option "shebang", type: :string, banner: "Specify a different shebang executable name than the default (usually 'ruby')"
+    method_option "shebang", type: :string, banner: "Specify a different shebang executable name than the default, usually 'ruby' (removed)"
     method_option "standalone", type: :array, lazy_default: [], banner: "Make a bundle that can work without the Bundler runtime"
-    method_option "system", type: :boolean, banner: "Install to the system location ($BUNDLE_PATH or $GEM_HOME) even if the bundle was previously installed somewhere else for this application"
+    method_option "system", type: :boolean, banner: "Install to the system location ($BUNDLE_PATH or $GEM_HOME) even if the bundle was previously installed somewhere else for this application (removed)"
     method_option "trust-policy", alias: "P", type: :string, banner: "Gem trust policy (like gem install -P). Must be one of #{Bundler.rubygems.security_policy_keys.join("|")}"
     method_option "target-rbconfig", type: :string, banner: "Path to rbconfig.rb for the deployment target platform"
-    method_option "without", type: :array, banner: "Exclude gems that are part of the specified named group."
-    method_option "with", type: :array, banner: "Include gems that are part of the specified named group."
+    method_option "without", type: :array, banner: "Exclude gems that are part of the specified named group (removed)."
+    method_option "with", type: :array, banner: "Include gems that are part of the specified named group (removed)."
     def install
       %w[clean deployment frozen no-prune path shebang without with].each do |option|
         remembered_flag_deprecation(option)
@@ -240,9 +276,16 @@ module Bundler
 
       remembered_flag_deprecation("deployment", negative: true)
 
+      if ARGV.include?("--binstubs")
+        removed_message = "The --binstubs option has been removed in favor of `bundle binstubs --all`"
+        raise InvalidOption, removed_message
+      end
+
       require_relative "cli/install"
+      options = self.options.dup
+      options["lockfile"] ||= ENV["BUNDLE_LOCKFILE"]
       Bundler.settings.temporary(no_install: false) do
-        Install.new(options.dup).run
+        Install.new(options).run
       end
     end
 
@@ -284,12 +327,11 @@ module Bundler
       Calling show with [GEM] will list the exact location of that gem on your machine.
     D
     method_option "paths", type: :boolean, banner: "List the paths of all gems that are required by your Gemfile."
-    method_option "outdated", type: :boolean, banner: "Show verbose output including whether gems are outdated."
+    method_option "outdated", type: :boolean, banner: "Show verbose output including whether gems are outdated (removed)."
     def show(gem_name = nil)
       if ARGV.include?("--outdated")
-        message = "the `--outdated` flag to `bundle show` will be removed in favor of `bundle show --verbose`"
         removed_message = "the `--outdated` flag to `bundle show` has been removed in favor of `bundle show --verbose`"
-        SharedHelpers.major_deprecation(2, message, removed_message: removed_message)
+        raise InvalidOption, removed_message
       end
       require_relative "cli/show"
       Show.new(options, gem_name).run
@@ -299,6 +341,7 @@ module Bundler
     method_option "name-only", type: :boolean, banner: "print only the gem names"
     method_option "only-group", type: :array, default: [], banner: "print gems from a given set of groups"
     method_option "without-group", type: :array, default: [], banner: "print all gems except from a given set of groups"
+    method_option "format", type: :string, banner: "format output ('json' is the only supported format)"
     method_option "paths", type: :boolean, banner: "print the path to each gem in the bundle"
     def list
       require_relative "cli/list"
@@ -322,7 +365,7 @@ module Bundler
       will create binstubs for all given gems.
     D
     method_option "force", type: :boolean, default: false, banner: "Overwrite existing binstubs if they exist"
-    method_option "path", type: :string, lazy_default: "bin", banner: "Binstub destination directory (default bin)"
+    method_option "path", type: :string, lazy_default: "bin", banner: "Binstub destination directory, `bin` by default (removed)"
     method_option "shebang", type: :string, banner: "Specify a different shebang executable name than the default (usually 'ruby')"
     method_option "standalone", type: :boolean, banner: "Make binstubs that can work without the Bundler runtime"
     method_option "all", type: :boolean, banner: "Install binstubs for all gems"
@@ -395,15 +438,15 @@ module Bundler
     end
 
     desc "cache [OPTIONS]", "Locks and then caches all of the gems into vendor/cache"
-    method_option "all", type: :boolean, default: Bundler.feature_flag.cache_all?, banner: "Include all sources (including path and git)."
+    method_option "all", type: :boolean, default: Bundler.settings[:cache_all], banner: "Include all sources (including path and git) (removed)."
     method_option "all-platforms", type: :boolean, banner: "Include gems for all platforms present in the lockfile, not only the current one"
     method_option "cache-path", type: :string, banner: "Specify a different cache path than the default (vendor/cache)."
     method_option "gemfile", type: :string, banner: "Use the specified gemfile instead of Gemfile"
     method_option "no-install", type: :boolean, banner: "Don't install the gems, only update the cache."
-    method_option "no-prune", type: :boolean, banner: "Don't remove stale gems from the cache."
-    method_option "path", type: :string, banner: "Specify a different path than the system default ($BUNDLE_PATH or $GEM_HOME).#{" Bundler will remember this value for future installs on this machine" unless Bundler.feature_flag.forget_cli_options?}"
+    method_option "no-prune", type: :boolean, banner: "Don't remove stale gems from the cache (removed)."
+    method_option "path", type: :string, banner: "Specify a different path than the system default, namely, $BUNDLE_PATH or $GEM_HOME (removed)."
     method_option "quiet", type: :boolean, banner: "Only output warnings and errors."
-    method_option "frozen", type: :boolean, banner: "Do not allow the Gemfile.lock to be updated after this bundle cache operation's install"
+    method_option "frozen", type: :boolean, banner: "Do not allow the Gemfile.lock to be updated after this bundle cache operation's install (removed)"
     long_desc <<-D
       The cache command will copy the .gem files for every gem in the bundle into the
       directory ./vendor/cache. If you then check that directory into your source
@@ -419,16 +462,12 @@ module Bundler
       end
 
       if flag_passed?("--path")
-        message =
-          "The `--path` flag is deprecated because its semantics are unclear. " \
-          "Use `bundle config cache_path` to configure the path of your cache of gems, " \
-          "and `bundle config path` to configure the path where your gems are installed, " \
-          "and stop using this flag"
         removed_message =
           "The `--path` flag has been removed because its semantics were unclear. " \
           "Use `bundle config cache_path` to configure the path of your cache of gems, " \
-          "and `bundle config path` to configure the path where your gems are installed."
-        SharedHelpers.major_deprecation 2, message, removed_message: removed_message
+          "and `bundle config path` to configure the path where your gems are installed, " \
+          "and stop using this flag"
+        raise InvalidOption, removed_message
       end
 
       require_relative "cli/cache"
@@ -438,7 +477,7 @@ module Bundler
     map aliases_for("cache")
 
     desc "exec [OPTIONS]", "Run the command in context of the bundle"
-    method_option :keep_file_descriptors, type: :boolean, default: true, banner: "Passes all file descriptors to the new processes. Default is true, and setting it to false is deprecated"
+    method_option :keep_file_descriptors, type: :boolean, default: true, banner: "Passes all file descriptors to the new processes. Default is true, and setting it to false is not permitted (removed)."
     method_option :gemfile, type: :string, required: false, banner: "Use the specified gemfile instead of Gemfile"
     long_desc <<-D
       Exec runs a command, providing it access to the gems in the bundle. While using
@@ -447,9 +486,8 @@ module Bundler
     D
     def exec(*args)
       if ARGV.include?("--no-keep-file-descriptors")
-        message = "The `--no-keep-file-descriptors` has been deprecated. `bundle exec` no longer mess with your file descriptors. Close them in the exec'd script if you need to"
         removed_message = "The `--no-keep-file-descriptors` has been removed. `bundle exec` no longer mess with your file descriptors. Close them in the exec'd script if you need to"
-        SharedHelpers.major_deprecation(2, message, removed_message: removed_message)
+        raise InvalidOption, removed_message
       end
 
       require_relative "cli/exec"
@@ -493,7 +531,7 @@ module Bundler
         build_info = " (#{BuildMetadata.timestamp} commit #{BuildMetadata.git_commit_sha})"
       end
 
-      if !cli_help && Bundler.feature_flag.bundler_4_mode?
+      if !cli_help
         Bundler.ui.info "#{Bundler.verbose_version}#{build_info}"
       else
         Bundler.ui.info "Bundler version #{Bundler.verbose_version}#{build_info}"
@@ -516,23 +554,9 @@ module Bundler
       end
     end
 
-    unless Bundler.feature_flag.bundler_4_mode?
-      desc "viz [OPTIONS]", "Generates a visual dependency graph", hide: true
-      long_desc <<-D
-        Viz generates a PNG file of the current Gemfile as a dependency graph.
-        Viz requires the ruby-graphviz gem (and its dependencies).
-        The associated gems must also be installed via 'bundle install'.
-      D
-      method_option :file, type: :string, default: "gem_graph", aliases: "-f", banner: "The name to use for the generated file. see format option"
-      method_option :format, type: :string, default: "png", aliases: "-F", banner: "This is output format option. Supported format is png, jpg, svg, dot ..."
-      method_option :requirements, type: :boolean, default: false, aliases: "-R", banner: "Set to show the version of each required dependency."
-      method_option :version, type: :boolean, default: false, aliases: "-v", banner: "Set to show each gem version."
-      method_option :without, type: :array, default: [], aliases: "-W", banner: "Exclude gems that are part of the specified named group."
-      def viz
-        SharedHelpers.major_deprecation 2, "The `viz` command has been renamed to `graph` and moved to a plugin. See https://github.com/rubygems/bundler-graph"
-        require_relative "cli/viz"
-        Viz.new(options.dup).run
-      end
+    desc "viz [OPTIONS]", "Generates a visual dependency graph", hide: true
+    def viz
+      SharedHelpers.feature_removed! "The `viz` command has been renamed to `graph` and moved to a plugin. See https://github.com/rubygems/bundler-graph"
     end
 
     desc "gem NAME [OPTIONS]", "Creates a skeleton for creating a rubygem"
@@ -542,7 +566,7 @@ module Bundler
     method_option :ext, type: :string, banner: "Generate the boilerplate for C extension code.", enum: EXTENSIONS
     method_option :git, type: :boolean, default: true, banner: "Initialize a git repo inside your library."
     method_option :mit, type: :boolean, banner: "Generate an MIT license file. Set a default with `bundle config set --global gem.mit true`."
-    method_option :rubocop, type: :boolean, banner: "Add rubocop to the generated Rakefile and gemspec. Set a default with `bundle config set --global gem.rubocop true`."
+    method_option :rubocop, type: :boolean, banner: "Add rubocop to the generated Rakefile and gemspec. Set a default with `bundle config set --global gem.rubocop true` (removed)."
     method_option :changelog, type: :boolean, banner: "Generate changelog file. Set a default with `bundle config set --global gem.changelog true`."
     method_option :test, type: :string, lazy_default: Bundler.settings["gem.test"] || "", aliases: "-t", banner: "Use the specified test framework for your library", enum: %w[rspec minitest test-unit], desc: "Generate a test directory for your library, either rspec, minitest or test-unit. Set a default with `bundle config set --global gem.test (rspec|minitest|test-unit)`."
     method_option :ci, type: :string, lazy_default: Bundler.settings["gem.ci"] || "", enum: %w[github gitlab circle], banner: "Generate CI configuration, either GitHub Actions, GitLab CI or CircleCI. Set a default with `bundle config set --global gem.ci (github|gitlab|circle)`"
@@ -552,6 +576,10 @@ module Bundler
 
     def gem(name)
       require_relative "cli/gem"
+
+      raise InvalidOption, "--rubocop has been removed, use --linter=rubocop" if ARGV.include?("--rubocop")
+      raise InvalidOption, "--no-rubocop has been removed, use --no-linter" if ARGV.include?("--no-rubocop")
+
       cmd_args = args + [self]
       cmd_args.unshift(options)
 
@@ -562,7 +590,7 @@ module Bundler
       File.expand_path("templates", __dir__)
     end
 
-    desc "clean [OPTIONS]", "Cleans up unused gems in your bundler directory", hide: true
+    desc "clean [OPTIONS]", "Cleans up unused gems in your bundler directory"
     method_option "dry-run", type: :boolean, default: false, banner: "Only print out changes, do not clean gems"
     method_option "force", type: :boolean, default: false, banner: "Forces cleaning up unused gems even if Bundler is configured to use globally installed gems. As a consequence, removes all system gems except for the ones in the current application."
     def clean
@@ -578,12 +606,8 @@ module Bundler
     end
 
     desc "inject GEM VERSION", "Add the named gem, with version requirements, to the resolved Gemfile", hide: true
-    method_option "source", type: :string, banner: "Install gem from the given source"
-    method_option "group", type: :string, banner: "Install gem into a bundler group"
-    def inject(name, version)
-      SharedHelpers.major_deprecation 2, "The `inject` command has been replaced by the `add` command"
-      require_relative "cli/inject"
-      Inject.new(options.dup, name, version).run
+    def inject(*)
+      SharedHelpers.feature_removed! "The `inject` command has been replaced by the `add` command"
     end
 
     desc "lock", "Creates a lockfile without installing"
@@ -637,7 +661,7 @@ module Bundler
       end
     end
 
-    if Bundler.feature_flag.plugins?
+    if Bundler.settings[:plugins]
       require_relative "cli/plugin"
       desc "plugin", "Manage the bundler plugins"
       subcommand "plugin", Plugin
@@ -671,18 +695,15 @@ module Bundler
       end
     end
 
-    def self.check_deprecated_ext_option(arguments)
-      # when deprecated version of `--ext` is called
-      # print out deprecation warning and pretend `--ext=c` was provided
-      if deprecated_ext_value?(arguments)
-        message = "Extensions can now be generated using C or Rust, so `--ext` with no arguments has been deprecated. Please select a language, e.g. `--ext=rust` to generate a Rust extension. This gem will now be generated as if `--ext=c` was used."
+    def self.check_invalid_ext_option(arguments)
+      # when invalid version of `--ext` is called
+      if invalid_ext_value?(arguments)
         removed_message = "Extensions can now be generated using C or Rust, so `--ext` with no arguments has been removed. Please select a language, e.g. `--ext=rust` to generate a Rust extension."
-        SharedHelpers.major_deprecation 2, message, removed_message: removed_message
-        arguments[arguments.index("--ext")] = "--ext=c"
+        raise InvalidOption, removed_message
       end
     end
 
-    def self.deprecated_ext_value?(arguments)
+    def self.invalid_ext_value?(arguments)
       index = arguments.index("--ext")
       next_argument = arguments[index + 1]
 
@@ -690,15 +711,15 @@ module Bundler
       # for example `bundle gem hello --ext c`
       return false if EXTENSIONS.include?(next_argument)
 
-      # deprecated call when --ext is called with no value in last position
+      # invalid call when --ext is called with no value in last position
       # for example `bundle gem hello_gem --ext`
       return true if next_argument.nil?
 
-      # deprecated call when --ext is followed by other parameter
+      # invalid call when --ext is followed by other parameter
       # for example `bundle gem --ext --no-ci hello_gem`
       return true if next_argument.start_with?("-")
 
-      # deprecated call when --ext is followed by gem name
+      # invalid call when --ext is followed by gem name
       # for example `bundle gem --ext hello_gem`
       return true if next_argument
 
@@ -712,6 +733,19 @@ module Bundler
       config[:current_command]
     end
 
+    def invoke_other_command(name)
+      _, _, config = @_initializer
+      original_command = config[:current_command]
+      command = self.class.all_commands[name]
+      config[:current_command] = command
+      send(name)
+    ensure
+      config[:current_command] = original_command
+    end
+
+    def current_command=(command)
+    end
+
     def print_command
       return unless Bundler.ui.debug?
       cmd = current_command
@@ -766,17 +800,12 @@ module Bundler
     end
 
     def print_remembered_flag_deprecation(flag_name, option_name, option_value)
-      message =
-        "The `#{flag_name}` flag is deprecated because it relies on being " \
-        "remembered across bundler invocations, which bundler will no longer " \
-        "do in future versions. Instead please use `bundle config set #{option_name} " \
-        "#{option_value}`, and stop using this flag"
       removed_message =
         "The `#{flag_name}` flag has been removed because it relied on being " \
-        "remembered across bundler invocations, which bundler will no longer " \
-        "do. Instead please use `bundle config set #{option_name} " \
-        "#{option_value}`, and stop using this flag"
-      Bundler::SharedHelpers.major_deprecation 2, message, removed_message: removed_message
+        "remembered across bundler invocations, which bundler no longer does. " \
+        "Instead please use `bundle config set #{option_name} #{option_value}`, " \
+        "and stop using this flag"
+      raise InvalidOption, removed_message
     end
 
     def flag_passed?(name)
